#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.hdlBackEnd as hdl
import revedaEditor.backend.libraryMethods as libm  # library view functions
import revedaEditor.backend.schBackEnd as scb  # import the backend
import revedaEditor.common.shapes as shp
import revedaEditor.fileio.symbolEncoder as se
import revedaEditor.gui.editorWindows as edw
import revedaEditor.gui.propertyDialogues as pdlg
import revedaEditor.gui.fileDialogues as fd
from PySide6.QtCore import Qt, QPoint
from PySide6.QtWidgets import QMainWindow, QDialog

import pathlib
import shutil
import json


def createSpiceView(
    parent: QMainWindow,
    importDlg: QDialog,
    libraryModel: edw.designLibrariesModel,
    importedSpiceObj: hdl.spiceC,
):
    """
    Create a new Spice view.

    Args:
        parent (QMainWindow): The parent window.
        importDlg (QDialog): The import dialog window.
        libraryModel (edw.designLibrariesModel): The model for the design libraries.
        importedSpiceObj (hdl.spiceC): The imported Spice object.

    Returns:
        tuple: A tuple containing the library item, cell item, and Spice item.
    """
    # Get the file path of the imported Spice file
    importedSpiceFilePathObj = pathlib.Path(importDlg.spiceFileEdit.text())
    # Get the selected library item
    libItem = libm.getLibItem(libraryModel, importDlg.libNamesCB.currentText())
    libItemRow = libItem.row()

    # Get the cell names in the selected library
    libCellNames = [
        libraryModel.item(libItemRow).child(i).cellName
        for i in range(libraryModel.item(libItemRow).rowCount())
    ]

    # Get the selected cell name
    cellName = importDlg.cellNamesCB.currentText().strip()

    # If the cell name is not in the library and is not empty, create a new cell
    if cellName not in libCellNames and cellName != "":
        scb.createCell(parent, libraryModel, libItem, cellName)

        # Get the cell item
    cellItem = libm.getCellItem(libItem, cellName)
    newSpiceFilePathObj = cellItem.data(Qt.UserRole + 2).joinpath(
        importedSpiceFilePathObj.name
    )
    # Create the Spice item view
    spiceItem = scb.createCellView(parent, importDlg.spiceViewName.text(), cellItem)
    # Create a temporary copy of the imported Spice file
    tempSpiceFilePathObj = importedSpiceFilePathObj.with_suffix(".tmp")
    shutil.copy(importedSpiceFilePathObj, tempSpiceFilePathObj)

    shutil.copy(tempSpiceFilePathObj, newSpiceFilePathObj)
    # Remove the temporary file
    tempSpiceFilePathObj.unlink()

    # Create a list of items to be stored in the Spice item data
    items = list()
    items.insert(0, {"cellView": "spice"})
    items.insert(1, {"filePath": str(newSpiceFilePathObj.name)})
    items.insert(2, {"subcktParams": importedSpiceObj.subcktParams})

    # Write the items to the Verilog-A item data file
    with spiceItem.data(Qt.UserRole + 2).open(mode="w") as f:
        json.dump(items, f, indent=4)

    # Return the tuple of library item, cell item, and Verilog-A item
    return ddef.viewItemTuple(libItem, cellItem, spiceItem)


def createVaView(
    parent: QMainWindow,
    importDlg: QDialog,
    libraryModel: edw.designLibrariesModel,
    importedVaObj: hdl.verilogaC,
) -> ddef.viewItemTuple:
    """
    Create a new Verilog-A view.

    Args:
        parent (QMainWindow): The parent window.
        importDlg (QDialog): The import dialog window.
        libraryModel (edw.designLibrariesModel): The model for the design libraries.
        importedVaObj (hdl.verilogaC): The imported Verilog-A object.

    Returns:
        tuple: A tuple containing the library item, cell item, and Verilog-A item.
    """
    # Get the file path of the imported Verilog-A file
    importedVaFilePathObj = pathlib.Path(importDlg.vaFileEdit.text())

    # Get the selected library item
    libItem = libm.getLibItem(libraryModel, importDlg.libNamesCB.currentText())
    libItemRow = libItem.row()

    # Get the cell names in the selected library
    libCellNames = [
        libraryModel.item(libItemRow).child(i).cellName
        for i in range(libraryModel.item(libItemRow).rowCount())
    ]

    # Get the selected cell name
    cellName = importDlg.cellNamesCB.currentText().strip()

    # If the cell name is not in the library and is not empty, create a new cell
    if cellName not in libCellNames and cellName != "":
        scb.createCell(parent, libraryModel, libItem, cellName)

    # Get the cell item
    cellItem = libm.getCellItem(libItem, cellName)

    # Generate the new file path for the Verilog-A file
    newVaFilePathObj = cellItem.data(Qt.UserRole + 2).joinpath(
        importedVaFilePathObj.name
    )

    # Create the Verilog-A item view
    vaItem = scb.createCellView(parent, importDlg.vaViewName.text(), cellItem)

    # Create a temporary copy of the imported Verilog-A file
    tempFilePathObj = importedVaFilePathObj.with_suffix(".temp")
    shutil.copy(importedVaFilePathObj, tempFilePathObj)

    # Copy the temporary file to the new file path
    shutil.copy(tempFilePathObj, newVaFilePathObj)

    # Remove the temporary file
    tempFilePathObj.unlink()

    # Create a list of items to be stored in the Verilog-A item data
    items = list()
    items.insert(0, {"cellView": "veriloga"})
    items.insert(1, {"filePath": str(newVaFilePathObj.name)})
    items.insert(2, {"vaModule": importedVaObj.vaModule})

    # Write the items to the Verilog-A item data file
    with vaItem.data(Qt.UserRole + 2).open(mode="w") as f:
        json.dump(items, f, indent=4)

    # Return the tuple of library item, cell item, and Verilog-A item
    return ddef.viewItemTuple(libItem, cellItem, vaItem)


def createSpiceSymbol(
    parent: QMainWindow,
    spiceItemTuple: ddef.viewItemTuple,
    libraryDict: dict,
    libraryBrowser: edw.libraryBrowser,
    importedSpiceObj: hdl.spiceC,
):
    symbolNameDlg = fd.createCellViewDialog(
        parent, libraryBrowser.libraryModel, spiceItemTuple.cellItem
    )
    symbolNameDlg.viewComboBox.setCurrentText("symbol")
    if symbolNameDlg.exec() == QDialog.Accepted:
        symbolViewName = symbolNameDlg.nameEdit.text().strip()
        symbolViewItem = scb.createCellView(
            parent, symbolViewName, spiceItemTuple.cellItem
        )
        newSpiceFilePathObj = spiceItemTuple.cellItem.data(Qt.UserRole + 2).joinpath(
            importedSpiceObj.pathObj.name
        )
        symbolWindow = edw.symbolEditor(
            symbolViewItem,
            libraryDict,
            libraryBrowser.libBrowserCont.designView,
        )
        symbolScene = symbolWindow.centralW.scene
        dlg = pdlg.symbolCreateDialog(parent)
        dlg.leftPinsEdit.setText(", ".join(importedSpiceObj.subcktParams["pins"]))

        if dlg.exec() == QDialog.Accepted:
            rectXDim, rectYDim = drawBaseSymbol(symbolScene, dlg)
            symbolFileLabel = symbolScene.labelDraw(
                QPoint(int(0.25 * rectXDim), int(-0.2 * rectYDim)),
                f"[@subcktName:subcktName=%:subcktName={importedSpiceObj.subcktParams['name']}]",
                "NLPLabel",
                "12",
                "Center",
                "R0",
                "Instance",
            )
            symbolFileLabel.labelVisible = False
            instParamNum = len(importedSpiceObj.subcktParams["params"])
            for index, (key, value) in enumerate(
                importedSpiceObj.subcktParams["params"].items()
            ):
                symbolScene.labelDraw(
                    QPoint(
                        int(rectXDim),
                        int(index * 0.2 * rectYDim / instParamNum),
                    ),
                    f"[@{key}:{key}=%:{key}={value}]",
                    "NLPLabel",
                    "12",
                    "Center",
                    "R0",
                    "Instance",
                )
            symbolScene.attributeList = list()
            symbolScene.attributeList.append(
                se.symbolAttribute("pinOrder", importedSpiceObj.pinOrder)
            )
            symbolScene.attributeList.append(
                se.symbolAttribute("incLine", f".INC {str(newSpiceFilePathObj)}\n")
            )

            symbolScene.attributeList.append(
                se.symbolAttribute("XyceSpiceNetlistLine", importedSpiceObj.netlistLine)
            )

            symbolWindow.show()
            symbolViewTuple = ddef.viewTuple(
                spiceItemTuple.libraryItem.libraryName,
                spiceItemTuple.cellItem.cellName,
                symbolViewItem.viewName,
            )
            symbolWindow.libraryView.openViews[symbolViewTuple] = symbolWindow


def createVaSymbol(
    parent: QMainWindow,
    vaItemTuple: ddef.viewItemTuple,
    libraryDict: dict,
    libraryBrowser: edw.libraryBrowser,
    importedVaObj: hdl.verilogaC,
) -> None:
    """
    Creates a symbol for a given view item in the library.

    Args:
        parent (QMainWindow): The parent window.
        vaItemTuple (ddef.viewItemTuple): The view item tuple.
        libraryDict (dict): The library dictionary.
        libraryBrowser (edw.libraryBrowser): The library browser.
        importedVaObj (hdl.verilogaC): The imported Veriloga object.

    Returns:
        None

    Raises:
        None
    """
    symbolNameDlg = fd.createCellViewDialog(
        parent, libraryBrowser.libraryModel, vaItemTuple.cellItem
    )
    symbolNameDlg.viewComboBox.setCurrentText("symbol")
    symbolNameDlg.nameEdit.setText("symbol")
    if symbolNameDlg.exec() == QDialog.Accepted:
        symbolViewName = symbolNameDlg.nameEdit.text().strip()
        symbolViewItem = scb.createCellView(
            parent, symbolViewName, vaItemTuple.cellItem
        )
        newVaFilePathObj = vaItemTuple.cellItem.data(Qt.UserRole + 2).joinpath(
            importedVaObj.pathObj.name
        )
        symbolWindow = edw.symbolEditor(
            symbolViewItem,
            libraryDict,
            libraryBrowser.libBrowserCont.designView,
        )
        symbolScene = symbolWindow.centralW.scene
        dlg = pdlg.symbolCreateDialog(parent)

        dlg.leftPinsEdit.setText(",".join(importedVaObj.inPins))
        dlg.rightPinsEdit.setText(",".join(importedVaObj.outPins))
        dlg.topPinsEdit.setText(",".join(importedVaObj.inoutPins))

        if dlg.exec() == QDialog.Accepted:
            rectXDim, rectYDim = drawBaseSymbol(symbolScene, dlg)
            # vaFileLabel = symbolScene.labelDraw(
            #     QPoint(int(0.25 * rectXDim), int(0.6 * rectYDim)),
            #     f"[@vaFile:vaFile=%:vaFile={str(newVaFilePathObj)}]",
            #     "NLPLabel",
            #     "12",
            #     "Center",
            #     "R0",
            #     "Instance",
            # )
            # vaFileLabel.labelVisible = False
            vaModuleLabel = symbolScene.labelDraw(
                QPoint(int(0.25 * rectXDim), int(0.8 * rectYDim)),
                f"[@vaModule:vaModule=%:vaModule={importedVaObj.vaModule}]",
                "NLPLabel",
                "12",
                "Center",
                "R0",
                "Instance",
            )
            vaModuleLabel.labelVisible = True
            # vaModelLabel = symbolScene.labelDraw(
            #     QPoint(int(0.25 * rectXDim), int(1 * rectYDim)),
            #     f"[@vaModel:vaModel=%:vaModel={importedVaObj.vaModule}Model]",
            #     "NLPLabel",
            #     "12",
            #     "Center",
            #     "R0",
            #     "Instance",
            # )
            # vaModelLabel.labelVisible = False

            instParamNum = len(importedVaObj.instanceParams)
            if instParamNum > 0:
                for index, (key, value) in enumerate(
                    importedVaObj.instanceParams.items()
                ):
                    symbolScene.labelDraw(
                        QPoint(
                            int(rectXDim),
                            int(index * 0.2 * rectYDim / instParamNum),
                        ),
                        f"[@{key}:{key}=%:{key}={value}]",
                        "NLPLabel",
                        "12",
                        "Center",
                        "R0",
                        "Instance",
                    )

            symbolScene.attributeList = list()  # empty attribute list
            if importedVaObj.modelParams:
                for key, value in importedVaObj.modelParams.items():
                    symbolScene.attributeList.append(se.symbolAttribute(key, value))
                symbolScene.attributeList.append(
                    se.symbolAttribute(
                        "XyceVerilogaNetlistLine", importedVaObj.netlistLine
                    )
                )

            modelParamsString = ", ".join(
                f"{key} = {value}" for key, value in importedVaObj.modelParams.items()
            )
            symbolScene.attributeList.append(
                se.symbolAttribute(
                    "vaModelLine",
                    f".MODEL {importedVaObj.vaModule}Model {importedVaObj.vaModule} {modelParamsString}",
                )
            )
            symbolScene.attributeList.append(
                se.symbolAttribute("vaHDLLine", f"*.HDL {str(importedVaObj.pathObj)}")
            )
            symbolScene.attributeList.append(
                se.symbolAttribute("pinOrder", importedVaObj.pinOrder)
            )
            symbolWindow.show()
            symbolViewTuple = ddef.viewTuple(
                vaItemTuple.libraryItem.libraryName,
                vaItemTuple.cellItem.cellName,
                "symbol",
            )
            symbolWindow.libraryView.openViews[symbolViewTuple] = symbolWindow


def drawBaseSymbol(symbolScene, dlg):
    leftPinNames = list(
        filter(
            None,
            [pinName.strip() for pinName in dlg.leftPinsEdit.text().split(",")],
        )
    )
    rightPinNames = list(
        filter(
            None,
            [pinName.strip() for pinName in dlg.rightPinsEdit.text().split(",")],
        )
    )
    topPinNames = list(
        filter(
            None,
            [pinName.strip() for pinName in dlg.topPinsEdit.text().split(",")],
        )
    )
    bottomPinNames = list(
        filter(
            None,
            [pinName.strip() for pinName in dlg.bottomPinsEdit.text().split(",")],
        )
    )
    stubLength = (
        int(float(dlg.stubLengthEdit.text().strip()))
        if dlg.stubLengthEdit.text()
        else 60
    )
    pinDistance = (
        int(float(dlg.pinDistanceEdit.text().strip()))
        if dlg.pinDistanceEdit.text()
        else 80
    )
    rectXDim = (max(len(topPinNames), len(bottomPinNames)) + 1) * pinDistance
    rectYDim = (max(len(leftPinNames), len(rightPinNames)) + 1) * pinDistance

    symbolScene.rectDraw(
        QPoint(0, 0),
        QPoint(rectXDim, rectYDim),
    )
    symbolScene.labelDraw(
        QPoint(int(0.25 * rectXDim), int(0.4 * rectYDim)),
        "[@cellName]",
        "NLPLabel",
        "12",
        "Center",
        "R0",
        "Instance",
    )
    symbolScene.labelDraw(
        QPoint(int(rectXDim), int(-0.1 * rectYDim)),
        "[@instName]",
        "NLPLabel",
        "12",
        "Center",
        "R0",
        "Instance",
    )

    leftPinLocs = [
        QPoint(-stubLength, (i + 1) * pinDistance) for i in range(len(leftPinNames))
    ]
    rightPinLocs = [
        QPoint(rectXDim + stubLength, (i + 1) * pinDistance)
        for i in range(len(rightPinNames))
    ]
    bottomPinLocs = [
        QPoint((i + 1) * pinDistance, rectYDim + stubLength)
        for i in range(len(bottomPinNames))
    ]
    topPinLocs = [
        QPoint((i + 1) * pinDistance, -stubLength) for i in range(len(topPinNames))
    ]
    for i, pinName in enumerate(leftPinNames):
        symbolScene.lineDraw(
            leftPinLocs[i],
            leftPinLocs[i] + QPoint(stubLength, 0),
        )
        symbolScene.addItem(
            shp.symbolPin(
                leftPinLocs[i],
                pinName,
                shp.symbolPin.pinDirs[0],
                shp.symbolPin.pinTypes[0],
            )
        )
    for i, pinName in enumerate(rightPinNames):
        symbolScene.lineDraw(
            rightPinLocs[i],
            rightPinLocs[i] + QPoint(-stubLength, 0),
        )
        symbolScene.addItem(
            shp.symbolPin(
                rightPinLocs[i],
                pinName,
                shp.symbolPin.pinDirs[1],
                shp.symbolPin.pinTypes[0],
            )
        )
    for i, pinName in enumerate(topPinNames):
        symbolScene.lineDraw(
            topPinLocs[i],
            topPinLocs[i] + QPoint(0, stubLength),
        )
        symbolScene.addItem(
            shp.symbolPin(
                topPinLocs[i],
                pinName,
                shp.symbolPin.pinDirs[2],
                shp.symbolPin.pinTypes[2],
            )
        )
    for i, pinName in enumerate(bottomPinNames):
        symbolScene.lineDraw(
            bottomPinLocs[i],
            bottomPinLocs[i] + QPoint(0, -stubLength),
        )
        symbolScene.addItem(
            shp.symbolPin(
                bottomPinLocs[i],
                pinName,
                shp.symbolPin.pinDirs[2],
                shp.symbolPin.pinTypes[1],
            )
        )

    return rectXDim, rectYDim



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import json
import inspect
import pdk.layoutLayers as laylyr
import revedaEditor.common.layoutShapes as lshp


class layoutEncoder(json.JSONEncoder):
    def default(self, item):
        match type(item):
            case lshp.layoutInstance:
                itemDict = {
                    "type": "Inst",
                    "lib": item.libraryName,
                    "cell": item.cellName,
                    "view": item.viewName,
                    "nam": item.instanceName,
                    "ic": item.counter,
                    "loc": (item.scenePos() - item.scene().origin).toTuple(),
                    "ang": item.angle,
                }
            case lshp.layoutRect:
                itemDict = {
                    "type": "Rect",
                    "tl": item.mapToScene(item.rect.topLeft()).toTuple(),
                    "br": item.mapToScene(item.rect.bottomRight()).toTuple(),
                    "ang": item.angle,
                    "ln": laylyr.pdkAllLayers.index(item.layer),
                }
            case lshp.layoutPath:
                itemDict = {
                    "type": "Path",
                    "dfl1": item.mapToScene(item.draftLine.p1()).toTuple(),
                    "dfl2": item.mapToScene(item.draftLine.p2()).toTuple(),
                    "ln": laylyr.pdkDrawingLayers.index(item.layer),
                    "w": item.width,
                    "se": item.startExtend,
                    "ee": item.endExtend,
                    "md": item.mode,
                    "nam": item.name,
                    "ang": item.angle,
                }
            case lshp.layoutViaArray:
                viaDict = {
                    "st": item.via.mapToScene(item.via.start).toTuple(),
                    "vdt": item.via.viaDefTuple.name,
                    "w": item.via.width,
                    "h": item.via.height,
                    "ang": item.angle,
                }
                itemDict = {
                    "type": "Via",
                    "st": item.mapToScene(item.start).toTuple(),
                    "via": viaDict,
                    "sp": item.spacing,
                    "xn": item.xnum,
                    "yn": item.ynum,
                }
            case lshp.layoutPin:
                itemDict = {
                    "type": "Pin",
                    "tl": item.mapToScene(item.rect.topLeft()).toTuple(),
                    "br": item.mapToScene(item.rect.bottomRight()).toTuple(),
                    "pn": item.pinName,
                    "pd": item.pinDir,
                    "pt": item.pinType,
                    "ln": laylyr.pdkPinLayers.index(item.layer),
                    "ang": item.angle,
                }
            case lshp.layoutLabel:
                itemDict = {
                    "type": "Label",
                    "st": item.mapToScene(item.start).toTuple(),
                    "lt": item.labelText,
                    "ff": item.fontFamily,
                    "fs": item.fontStyle,
                    "fh": item.fontHeight,
                    "la": item.labelAlign,
                    "lo": item.labelOrient,
                    "ln": laylyr.pdkTextLayers.index(item.layer),
                    "ang": item.angle,
                }
            case lshp.layoutPolygon:
                pointsList = [item.mapToScene(point).toTuple() for point in item.points]
                itemDict = {
                    "type": "Polygon",
                    "ps": pointsList,
                    "ln": laylyr.pdkAllLayers.index(item.layer),
                    "ang": item.angle,
                }
            case lshp.layoutRuler:
                itemDict = {
                    "type": "Ruler",
                    "dfl1": item.mapToScene(item.draftLine.p1()).toTuple(),
                    "dfl2": item.mapToScene(item.draftLine.p2()).toTuple(),
                    "md": item.mode,
                    "ang": item.angle,
                }
            case _:  # now check super class types:
                match item.__class__.__bases__[0]:
                    case lshp.layoutPcell:
                        init_args = inspect.signature(
                            item.__class__.__init__
                        ).parameters
                        args_used = [param for param in init_args if (param != "self")]

                        argDict = {arg: getattr(item, arg) for arg in args_used}
                        # print(argDict)
                        itemDict = {
                            "type": "Pcell",
                            "lib": item.libraryName,
                            "cell": item.cellName,
                            "view": item.viewName,
                            "nam": item.instanceName,
                            "ic": item.counter,
                            "loc": (item.scenePos() - item.scene().origin).toTuple(),
                            "ang": item.angle,
                            "params": argDict,
                        }
        return itemDict



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import time
import logging


def debug_decorator(func):
    def wrapper(*args, **kwargs):
        print(f"Calling {func.__name__} called with {args} and kwargs: {kwargs}")
        return func(*args, **kwargs)

    return wrapper


def timer(func):
    def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        duration = (end_time - start_time) * 1000
        logging.debug(f"  Took {duration:0.4f} ms")
        return result

    return wrapper



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#   #
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#   #
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#   #
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)

from PySide6.QtCore import (
    Qt,
)
from PySide6.QtGui import QStandardItemModel

import revedaEditor.backend.schBackEnd as scb


def getLibItem(libraryModel: QStandardItemModel, libName: str) -> scb.libraryItem:
    libItem = [
        item
        for item in libraryModel.findItems(libName)
        if item.data(Qt.UserRole + 1) == "library"
    ][0]
    return libItem


def getCellItem(libItem: scb.libraryItem, cellNameInp: str) -> scb.cellItem:
    cellItems = [
        libItem.child(i)
        for i in range(libItem.rowCount())
        if libItem.child(i).cellName == cellNameInp
    ]
    if cellItems:
        return cellItems[0]


def getViewItem(cellItem: scb.cellItem, viewNameInp: str) -> scb.viewItem:
    if cellItem is not None:
        viewItems = [
            cellItem.child(i)
            for i in range(cellItem.rowCount())
            if cellItem.child(i).text() == viewNameInp
        ]
    if viewItems:
        return viewItems[0]


def findViewItem(libraryModel, libName: str, cellName: str, viewName: str):
    libItem = getLibItem(libraryModel, libName)
    cellItem = getCellItem(libItem, cellName)
    return getViewItem(cellItem, viewName)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import gdstk
import revedaEditor.common.layoutShapes as lshp
import pathlib
import inspect


class gdsExporter:
    def __init__(self, cellname: str, items: list, outputFileObj: pathlib.Path):
        self._cellname = cellname
        self._items = items
        self._outputFileObj = outputFileObj
        self._libraryName = None
        self._unit = 1e-6
        self._precision = 1e-9
        self._topCell = None
        self._cellNamesSet = set()

    def gds_export(self):
        self._outputFileObj.parent.mkdir(parents=True, exist_ok=True)
        lib = gdstk.Library(unit=self._unit, precision=self._precision)
        self._topCell = lib.new_cell(self._cellname)  # top Cell
        for item in self._items:
            self.createCells(lib, item, self._topCell)
        lib.write_gds(self._outputFileObj)

    def createCells(
        self, library: gdstk.Library, item: lshp.layoutShape, parentCell: gdstk.Cell
    ):
        match type(item):
            case lshp.layoutInstance:  # recursive search under a layout cell
                if item.cellName not in self._cellNamesSet:
                    cellGDSName = f"{item.libraryName}_{item.cellName}_{item.viewName}"
                    library.new_cell(cellGDSName)
                    self._cellNamesSet.add(cellGDSName)
                    for shape in item.shapes:
                        self.createCells(library, shape, library[cellGDSName])
                ref = gdstk.Reference(
                    library[cellGDSName],
                    item.pos().toPoint().toTuple(),
                    rotation=item.angle,
                )
                parentCell.add(ref)
            case lshp.layoutRect:
                rect = gdstk.rectangle(
                    corner1=item.start.toTuple(),
                    corner2=item.end.toTuple(),
                    layer=item.layer.gdsLayer,
                    datatype=item.layer.datatype,
                )
                parentCell.add(rect)
            case lshp.layoutPath:
                path = gdstk.FlexPath(
                    points=[
                        item.draftLine.p1().toTuple(),
                        item.draftLine.p2().toTuple(),
                    ],
                    width=item.width,
                    ends=(item.startExtend, item.endExtend),
                    simple_path=True,
                    layer=item.layer.gdsLayer,
                    datatype=item.layer.datatype,
                )
                parentCell.add(path)
            case lshp.layoutLabel:
                label = gdstk.Label(
                    text=item.labelText,
                    origin=item.start.toTuple(),
                    rotation=item.angle,
                    layer=item.layer.gdsLayer,
                )
                parentCell.add(label)
            case lshp.layoutPin:
                pin = gdstk.rectangle(
                    corner1=item.start.toTuple(),
                    corner2=item.end.toTuple(),
                    layer=item.layer.gdsLayer,
                    datatype=item.layer.datatype,
                )
                parentCell.add(pin)
            case lshp.layoutPolygon:
                points = [point.toTuple() for point in item.points]
                polygon = gdstk.Polygon(
                    points=points,
                    layer=item.layer.gdsLayer,
                    datatype=item.layer.datatype,
                )
                parentCell.add(polygon)
            case lshp.layoutViaArray:
                viaName = f"via_{item.via.width}_{item.via.height}_{item.via.layer.name}_{item.via.layer.purpose}"
                if viaName not in self._cellNamesSet:
                    viaCell = library.new_cell(
                        f"via_{item.via.width}_{item.via.height}_"
                        f"{item.via.layer.name}_{item.via.layer.purpose}"
                    )
                    via = gdstk.rectangle(
                        item.mapToScene(item.via.rect.topLeft()).toTuple(),
                        item.mapToScene(item.via.rect.bottomRight()).toTuple(),
                        layer=item.via.layer.gdsLayer,
                        datatype=item.via.layer.datatype,
                    )
                    self._cellNamesSet.add(viaCell.name)
                    viaCell.add(via)
                viaCell = library[viaName]
                viaArray = gdstk.Reference(
                    cell=viaCell,
                    origin=item.start.toTuple(),
                    columns=item.xnum,
                    rows=item.ynum,
                    spacing=(item.spacing + item.width, item.spacing + item.height),
                )
                parentCell.add(viaArray)
            case _:  # now check super class types:
                match item.__class__.__bases__[0]:
                    case lshp.layoutPcell:
                        pcellParamDict = gdsExporter.extractPcellInstanceParameters(
                            item
                        )
                        pcellNameSuffix = "_".join(
                            [f"{key}_{value}" for key, value in pcellParamDict.items()]
                        ).replace(".", "p")
                        pcellName = (
                            f"{item.libraryName}_{type(item).__name__}"
                            f"_{pcellNameSuffix}"
                        )
                        if pcellName not in self._cellNamesSet:
                            library.new_cell(pcellName)
                            self._cellNamesSet.add(pcellName)
                            for shape in item.shapes:
                                self.createCells(library, shape, library[pcellName])
                        ref = gdstk.Reference(
                            library[pcellName],
                            item.pos().toPoint().toTuple(),
                            rotation=item.angle,
                        )
                        parentCell.add(ref)

    @staticmethod
    def extractPcellInstanceParameters(instance: lshp.layoutPcell) -> dict:
        initArgs = inspect.signature(instance.__class__.__init__).parameters
        argsUsed = [
            param for param in initArgs if (param != "self" and param != "snapTuple")
        ]
        argDict = {arg: getattr(instance, arg) for arg in argsUsed}
        return argDict

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, value):
        self._unit = value

    @property
    def precision(self):
        return self._precision

    @precision.setter
    def precision(self, value):
        self._precision = value



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#


# shape class definition for symbol editor.
# base class for all shapes: rectangle, circle, line
import itertools
import math
from PySide6.QtCore import QPoint, QRect, QRectF, Qt, QPointF, QLineF
from PySide6.QtGui import (
    QPen,
    QBrush,
    QColor,
    QTransform,
    QPixmap,
    QBitmap,
    QFontMetrics,
    QFont,
    QTextOption,
    QFontDatabase,
    QPainterPath,
    QPolygonF,
)
from PySide6.QtWidgets import (
    QGraphicsItem,
    QGraphicsSceneMouseEvent,
    QGraphicsSceneHoverEvent,
)

import revedaEditor.backend.dataDefinitions as ddef


class layoutShape(QGraphicsItem):
    def __init__(self) -> None:
        super().__init__()
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
        self.setFlag(QGraphicsItem.ItemIsFocusable, True)
        self.setAcceptHoverEvents(True)
        self._pen = None
        self._brush = None
        self._angle = 0  # rotation angle
        self._stretch: bool = False
        self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)

    def __repr__(self):
        return "layoutShape()"

    def itemChange(self, change, value):
        return super().itemChange(change, value)

    @property
    def pen(self):
        return self._pen

    @pen.setter
    def pen(self, value: QPen):
        if isinstance(value, QPen):
            self._pen = value

    @property
    def brush(self):
        return self._brush

    @brush.setter
    def brush(self, value: QBrush):
        if isinstance(value, QBrush):
            self._brush = value

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, value):
        self._angle = value
        self.prepareGeometryChange()
        self.setRotation(value)  # self.update(self.boundingRect())

    @property
    def stretch(self):
        return self._stretch

    @stretch.setter
    def stretch(self, value: bool):
        self._stretch = value

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        if self.scene().editModes.changeOrigin or self.scene().drawMode:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)
        else:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def sceneEvent(self, event):
        """
        Do not propagate event if shape needs to keep still.
        """
        if self.scene() and (
            self.scene().editModes.changeOrigin or self.scene().drawMode
        ):
            return False
        else:
            super().sceneEvent(event)
            return True

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)  # self.setSelected(False)

    def hoverEnterEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        self.setCursor(Qt.ArrowCursor)
        self.setOpacity(0.75)
        self.setFocus()
        super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        super().hoverLeaveEvent(event)
        self.setCursor(Qt.CrossCursor)
        self.setOpacity(1)
        self.clearFocus()

    def contextMenuEvent(self, event):
        self.scene().itemContextMenu.exec_(event.screenPos())


class layoutRect(layoutShape):
    sides = ["Left", "Right", "Top", "Bottom"]

    def __init__(
        self,
        start: QPoint,
        end: QPoint,
        layer: ddef.layLayer,
    ):
        super().__init__()
        self._rect = QRectF(start, end).normalized()
        self._start = self._rect.topLeft()
        self._end = self._rect.bottomRight()
        self._layer = layer
        if self._layer.selectable:
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
            self.setFlag(QGraphicsItem.ItemIsFocusable, True)
        else:
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)
            self.setFlag(QGraphicsItem.ItemIsFocusable, False)
        self._stretch = False
        self._stretchSide = None
        self._stretchPen = QPen(QColor("red"), self._layer.pwidth, Qt.SolidLine)
        self._definePensBrushes()

    def _definePensBrushes(self):
        self._pen = QPen(self._layer.pcolor, self._layer.pwidth, self._layer.pstyle)
        self._bitmap = QBitmap.fromImage(
            QPixmap(self._layer.btexture).scaled(10, 10).toImage()
        )
        self._brush = QBrush(self._layer.bcolor, self._bitmap)
        self._selectedPen = QPen(QColor("yellow"), self._layer.pwidth, Qt.DashLine)
        self._selectedBrush = QBrush(QColor("yellow"), self._bitmap)

    def __repr__(self):
        return f"layoutRect({self._start}, {self._end}, {self._layer})"

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(self._selectedPen)
            painter.setBrush(self._selectedBrush)
            if self.stretch:
                painter.setPen(self._stretchPen)
                if self._stretchSide == layoutRect.sides[0]:
                    painter.drawLine(self.rect.topLeft(), self.rect.bottomLeft())
                elif self._stretchSide == layoutRect.sides[1]:
                    painter.drawLine(self.rect.topRight(), self.rect.bottomRight())
                elif self._stretchSide == layoutRect.sides[2]:
                    painter.drawLine(self.rect.topLeft(), self.rect.topRight())
                elif self._stretchSide == layoutRect.sides[3]:
                    painter.drawLine(self.rect.bottomLeft(), self.rect.bottomRight())
        else:
            painter.setPen(self._pen)
            painter.setBrush(self._brush)
        painter.drawRect(self._rect)

    def boundingRect(self):
        return self._rect.normalized().adjusted(-2, -2, 2, 2)

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, rect: QRect):
        self.prepareGeometryChange()
        self._rect = rect

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start: QPoint):
        self.prepareGeometryChange()
        self._rect = QRectF(start, self.end).normalized()
        self._start = start

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end: QPoint):
        self.prepareGeometryChange()
        self._rect = QRectF(self.start, end).normalized()
        self._end = end

    @property
    def centre(self):
        return QPoint(
            int(self._rect.x() + self._rect.width() / 2),
            int(self._rect.y() + self._rect.height() / 2),
        )

    @property
    def height(self):
        return self.rect.height()

    @height.setter
    def height(self, height: int):
        self.prepareGeometryChange()
        self._rect.setHeight(height)

    @property
    def width(self):
        return self.rect.width()

    @width.setter
    def width(self, width):
        self.prepareGeometryChange()
        self.rect.setWidth(width)

    @property
    def left(self):
        return self.rect.left()

    @left.setter
    def left(self, left: int):
        self.rect.setLeft(left)

    @property
    def right(self):
        return self.rect.right()

    @right.setter
    def right(self, right: int):
        self.prepareGeometryChange()
        self.rect.setRight(right)

    @property
    def top(self):
        return self.rect.top()

    @top.setter
    def top(self, top: int):
        self.prepareGeometryChange()
        self.rect.setTop(top)

    @property
    def bottom(self):
        return self.rect.bottom()

    @bottom.setter
    def bottom(self, bottom: int):
        self.prepareGeometryChange()
        self.rect.setBottom(bottom)

    @property
    def origin(self):
        return self.rect.bottomLeft()

    @property
    def stretchSide(self):
        return self._stretchSide

    @stretchSide.setter
    def stretchSide(self, value: str):
        self.prepareGeometryChange()
        self._stretchSide = value

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, value):
        self.prepareGeometryChange()
        self._layer = value
        self._definePensBrushes()

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        if self._layer.selectable:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
            eventPos = event.pos().toPoint()
            if self._stretch:
                self.setFlag(QGraphicsItem.ItemIsMovable, False)
                if eventPos.x() == self._rect.left():
                    if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                        self.setCursor(Qt.SizeHorCursor)
                        self._stretchSide = layoutRect.sides[0]
                elif eventPos.x() == self._rect.right():
                    if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                        self.setCursor(Qt.SizeHorCursor)
                        self._stretchSide = layoutRect.sides[1]
                elif eventPos.y() == self._rect.top():
                    if self._rect.left() <= eventPos.x() <= self._rect.right():
                        self.setCursor(Qt.SizeVerCursor)
                        self._stretchSide = layoutRect.sides[2]
                elif eventPos.y() == self._rect.bottom():
                    if self._rect.left() <= eventPos.x() <= self._rect.right():
                        self.setCursor(Qt.SizeVerCursor)
                        self._stretchSide = layoutRect.sides[3]
        else:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        eventPos = event.pos().toPoint()
        if self.stretch:
            self.prepareGeometryChange()
            if self.stretchSide == layoutRect.sides[0]:
                self.setCursor(Qt.SizeHorCursor)
                self.rect.setLeft(eventPos.x())
            elif self.stretchSide == layoutRect.sides[1]:
                self.setCursor(Qt.SizeHorCursor)
                self.rect.setRight(eventPos.x() - int(self._pen.width() / 2))
            elif self.stretchSide == layoutRect.sides[2]:
                self.setCursor(Qt.SizeVerCursor)
                self.rect.setTop(eventPos.y())
            elif self.stretchSide == layoutRect.sides[3]:
                self.setCursor(Qt.SizeVerCursor)
                self.rect.setBottom(eventPos.y() - int(self._pen.width() / 2))
            self.update()
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        super().mouseReleaseEvent(event)
        if self.stretch:
            self._stretch = False
            self._stretchSide = None
            self.setCursor(Qt.ArrowCursor)


class layoutInstance(layoutShape):
    def __init__(self, shapes: list[layoutShape]):
        super().__init__()
        # List of shapes in the symbol
        self._shapes = shapes
        # Flag to indicate if the symbol is in draft mode
        self._draft = False
        # Name of the library
        self._libraryName: str = ""
        # Name of the cell
        self._cellName: str = ""
        # Name of the view
        self._viewName: str = ""
        # Name of the instance
        self._instanceName = ""
        # Pen used for selection
        self._selectedPen = QPen(QColor("yellow"), 1, Qt.DashLine)
        # Set the shapes for the symbol
        self.setShapes()
        # Enable child event filtering for filters and handles
        self.setFiltersChildEvents(True)
        self.setHandlesChildEvents(True)
        # Enable flag to indicate that the item contains children in shape
        self.setFlag(QGraphicsItem.ItemContainsChildrenInShape, True)
        # Set the top left position of the symbol
        self._start = self.childrenBoundingRect().topLeft()

    def setShapes(self):
        for item in self._shapes:
            item.setFlag(QGraphicsItem.ItemIsSelectable, False)
            item.setFlag(QGraphicsItem.ItemStacksBehindParent, True)
            item.setParentItem(self)

    def removeShapes(self):
        self.prepareGeometryChange()
        for item in self._shapes:
            item.setParentItem(None)
            del item
        self._shapes = list()

    def __repr__(self):
        return f"layoutInstance({self._shapes})"

    def boundingRect(self):
        return self.childrenBoundingRect().normalized().adjusted(-2, -2, 2, 2)

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(self._selectedPen)
            painter.drawRect(self.childrenBoundingRect())

    @property
    def libraryName(self):
        return self._libraryName

    @libraryName.setter
    def libraryName(self, value):
        self._libraryName = value

    @property
    def cellName(self):
        return self._cellName

    @cellName.setter
    def cellName(self, value: str):
        self._cellName = value

    @property
    def viewName(self):
        return self._viewName

    @viewName.setter
    def viewName(self, value: str):
        self._viewName = value

    @property
    def instanceName(self):
        return self._instanceName

    @instanceName.setter
    def instanceName(self, value: str):
        assert isinstance(value, str)
        self._instanceName = value

    @property
    def shapes(self):
        return self._shapes

    @shapes.setter
    def shapes(self, value: list[layoutShape]):
        self.removeShapes()
        self._shapes = value
        self.setShapes()

    @property
    def start(self):
        return self._start.toPoint()

    def addShape(self, shape: layoutShape):
        self._shapes.append(shape)
        shape.setParentItem(self)


class layoutPcell(layoutInstance):
    def __init__(self, shapes: list):
        super().__init__(shapes)

    def __repr__(self):
        return f"layoutPcell({self._shapes}"


class layoutLine(layoutShape):
    def __init__(
        self,
        draftLine: QLineF,
        layer: ddef.layLayer,
        width: float = 1.0,
    ):
        super().__init__()
        self._draftLine = draftLine
        self._layer = layer
        self._width = width
        self._pen = QPen(self._layer.pcolor, self._layer.pwidth, self._layer.pstyle)
        self._selectedPen = QPen(QColor("yellow"), self._layer.pwidth, Qt.DashLine)
        self._rect = (
            QRectF(self._draftLine.p1(), self._draftLine.p2())
            .normalized()
            .adjusted(-2, -2, 2, 2)
        )

    def __repr__(self):
        return f"layoutLine({self._draftLine}, {self._layer}, {self._width})"

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(self._selectedPen)
        else:
            painter.setPen(self._pen)
        painter.drawLine(self._draftLine)

    def boundingRect(self):
        return self._rect


# "layer", "name", "mode", "width", "startExtend", "endExtend"


class layoutPath(layoutShape):
    def __init__(
        self,
        draftLine: QLineF,
        layer: ddef.layLayer,
        width: float = 1.0,
        startExtend: int = 0,
        endExtend: int = 0,
        mode: int = 0,
    ):
        """
        Initialize the class instance.

        Args:
            draftLine (QLineF): The draft line.
            layer (ddef.layLayer): The layer.
            width (float, optional): The width. Defaults to 1.0.
            startExtend (int, optional): The start extend. Defaults to 0.
            endExtend (int, optional): The end extend. Defaults to 0.
            mode (int, optional): The mode. Defaults to 0.
        """
        super().__init__()
        self.start = None
        self._draftLine = draftLine
        self._startExtend = startExtend
        self._endExtend = endExtend
        self._width = width
        self._layer = layer
        self._mode = mode
        self._name = ""
        self._stretch = False
        self._stretchSide = None
        self._definePensBrushes()
        self._rect = QRectF(0, 0, 0, 0)
        self._angle = 0
        self._rectCorners(self._draftLine.angle())

    def _definePensBrushes(self):
        self._pen = QPen(self._layer.pcolor, self._layer.pwidth, self._layer.pstyle)
        self._bitmap = QBitmap.fromImage(
            QPixmap(self._layer.btexture).scaled(10, 10).toImage()
        )
        self._brush = QBrush(self._layer.bcolor, self._bitmap)
        self._selectedPen = QPen(QColor("yellow"), self._layer.pwidth, Qt.DashLine)
        self._selectedBrush = QBrush(QColor("yellow"), self._bitmap)
        self._stretchPen = QPen(QColor("red"), self._layer.pwidth, Qt.SolidLine)
        self._stretchBrush = QBrush(QColor("red"), self._bitmap)

    def __repr__(self):
        return (
            f"layoutPath({self._draftLine}, {self._layer}"
            f"{self._width}, {self._startExtend}, {self._endExtend}, {self._mode})"
        )

    def _rectCorners(self, angle: float):
        match self._mode:
            case 0:  # manhattan
                self._createManhattanPath(angle)
            case 1:  # diagonal
                self._createDiagonalPath(angle)
            case 2:
                self._createAnyAnglePath(angle)
            case 3:
                self._createHorizontalPath(angle)
            case 4:
                self._createVerticalPath(angle)
        self._draftLine.setAngle(0)
        self._rect = self._extractRect()
        self.setTransformOriginPoint(self.draftLine.p1())
        self.setRotation(-self._angle)

    def _createManhattanPath(self, angle: float) -> None:
        """
        Creates a Manhattan path based on the given angle.

        :param angle: The angle in degrees.
        :type angle: float

        :return: None
        """
        self._angle = 90 * math.floor(((angle + 45) % 360) / 90)

    def _createDiagonalPath(self, angle: float) -> None:
        """
        Creates a manhattan or diagonal path based on the given angle.
        Parameters:
            angle (float): The angle in degrees.
        Returns:
            None
        """
        self._angle = 45 * math.floor(((angle + 22.5) % 360) / 45)

    def _createAnyAnglePath(self, angle: float) -> None:
        self._angle = angle

    def _createHorizontalPath(self, angle: float) -> None:
        self._angle = 180 * math.floor(((angle + 90) % 360) / 180)

    def _createVerticalPath(self, angle: float) -> None:
        angle = angle % 360
        if 0 <= angle < 180:
            self._angle = 90
        else:
            self._angle = 270

    def _extractRect(self):
        direction = self._draftLine.p2() - self._draftLine.p1()
        if direction == QPoint(0, 0):  # when the mouse pressed first time
            rect = (
                QRectF(self._draftLine.p1(), self._draftLine.p2())
                .adjusted(-2, -2, 2, 2)
                .normalized()
            )
        else:
            direction /= direction.manhattanLength()
            perpendicular = QPointF(-direction.y(), direction.x())
            point1 = (
                self._draftLine.p1()
                + perpendicular * self._width * 0.5
                - direction * self._startExtend
            ).toPoint()
            point2 = (
                self._draftLine.p2()
                - perpendicular * self._width * 0.5
                + direction * self._endExtend
            ).toPoint()
            rect = QRectF(point1, point2).normalized()
        return rect

    def paint(self, painter, option, widget):
        if self.isSelected():
            if self._stretch:
                painter.setPen(self._stretchPen)
                painter.setBrush(self._stretchBrush)
            else:
                painter.setPen(self._selectedPen)
                painter.setBrush(self._selectedBrush)
        else:
            painter.setPen(self._pen)
            painter.setBrush(self._brush)
        painter.drawLine(self._draftLine)
        painter.drawRect(self._rect)

    def boundingRect(self) -> QRectF:
        return self._rect.adjusted(-2, 2, 2, 2)

    @property
    def draftLine(self):
        return self._draftLine

    @draftLine.setter
    def draftLine(self, line: QLineF):
        self.prepareGeometryChange()
        self._draftLine = line
        angle = self._draftLine.angle()
        self._rectCorners(angle)

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width: float):
        self._width = width
        self.prepareGeometryChange()
        self._rectCorners(self._angle)

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value: int):
        self.prepareGeometryChange()
        self._mode = value
        self._rectCorners(self._angle)

    @property
    def stretchSide(self):
        return self._stretchSide

    @stretchSide.setter
    def stretchSide(self, value: str):
        self.prepareGeometryChange()
        self._stretchSide = value

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, value):
        self._layer = value
        self._definePensBrushes()

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def startExtend(self) -> int:
        return self._startExtend

    @startExtend.setter
    def startExtend(self, value: str):
        self.prepareGeometryChange()
        self._startExtend = value
        self._rect = self._extractRect()

    @property
    def endExtend(self) -> int:
        return self._endExtend

    @endExtend.setter
    def endExtend(self, value: str):
        self.prepareGeometryChange()
        self._endExtend = value
        self._rect = self._extractRect()

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, value: float):
        self._angle = value
        self.prepareGeometryChange()
        self._rect = self._extractRect()
        self.setTransformOriginPoint(self.draftLine.p1())
        self.setRotation(-self._angle)

    @property
    def sceneEndPoints(self):
        return [
            self.mapToScene(self._draftLine.p1()).toPoint(),
            self.mapToScene(self._draftLine.p2()).toPoint(),
        ]

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        if self._layer.selectable:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
            eventPos = event.pos().toPoint()
            if self._stretch:
                if (
                    eventPos - self._draftLine.p1().toPoint()
                ).manhattanLength() <= self.scene().snapDistance:
                    self._stretchSide = "p1"
                    self.setCursor(Qt.SizeHorCursor)
                elif (
                    eventPos - self._draftLine.p2().toPoint()
                ).manhattanLength() <= self.scene().snapDistance:
                    self._stretchSide = "p2"
                    self.setCursor(Qt.SizeHorCursor)
                self.scene().stretchPath(self, self._stretchSide)
        else:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        eventPos = event.pos().toPoint()
        # if self.stretch:
        #     self.prepareGeometryChange()
        #     if self._stretchSide == "p1":
        #         self.draftLine.setP1(eventPos)
        #     elif self._stretchSide == "p2":
        #         self.draftLine.setP2(eventPos)
        #     self._rectCorners(self.angle)
        # else:
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        #     self.setFlag(QGraphicsItem.ItemIsMovable, True)
        super().mouseReleaseEvent(event)

    #     if self.stretch:
    #         self._stretch = False
    #         self._stretchSide = None
    #         self.setCursor(Qt.ArrowCursor)


class layoutRuler(layoutShape):
    def __init__(
        self,
        draftLine: QLineF,
        width: float,
        tickGap: float,
        tickLength: int,
        tickFont: QFont,
        mode: int = 0,
    ):
        """
        Initialize the TickLine object.

        Args:
            draftLine (QLineF): The draft line.
            width (float): The width of the line.
            tickGap (float): The gap between ticks.
            tickLength (int): The length of the ticks.
            tickFont (QFont): The font for tick labels.
            mode (int, optional): The mode. Defaults to 0.
        """
        super().__init__()

        self._draftLine = draftLine
        self._width = width
        self._tickGap = tickGap
        self._tickLength = tickLength
        self._mode = mode
        self._angle = 0
        self._rect = QRect(0, 0, 0, 0)
        penColour = QColor(255, 255, 40)
        # penColour.setAlpha(128)
        self._pen = QPen(penColour, self._width, Qt.SolidLine)
        self._selectedPen = QPen(Qt.red, self._width + 1, Qt.SolidLine)
        # self._pen.setCosmetic(True)
        self._tickTuples = list()
        self._tickFont = tickFont
        self._determineAngle(self._draftLine.angle())
        self._fm = QFontMetrics(self._tickFont)
        # Enable child event filtering for filters and handles
        self.setFiltersChildEvents(True)
        self.setHandlesChildEvents(True)
        # Enable flag to indicate that the item contains children in shape
        self.setFlag(QGraphicsItem.ItemContainsChildrenInShape, True)
        self._createRulerTicks()
        # self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)
        self.update(self.boundingRect())

    def __repr__(self):
        return (
            f"layoutRuler({self._draftLine}, {self._width}, {self._tickGap}, "
            f"{self._tickLength}, {self._tickFont}, {self._mode})"
        )

    def _determineAngle(self, angle: float):
        match self._mode:
            case 0:  # manhattan
                self._createManhattanRuler(angle)
            case 1:  # diagonal
                self._createDiagonalRuler(angle)
            case 2:
                self._createAnyAngleRuler(angle)
        self._draftLine.setAngle(0)
        self.setTransformOriginPoint(self.draftLine.p1())
        self.setRotation(-self._angle)

    def _createManhattanRuler(self, angle):
        if 0 <= angle <= 45 or 360 > angle > 315:
            self._angle = 0
        elif 45 < angle <= 135:
            self._angle = 90
        elif 135 < angle <= 225:
            self._angle = 180
        elif 225 < angle <= 315:
            self._angle = 270

    def _createDiagonalRuler(self, angle):
        if 0 <= angle <= 22.5 or 360 > angle > 337.5:
            self._angle = 0
        elif 22.5 < angle <= 67.5:
            self._angle = 45
        elif 67.5 < angle <= 112.5:
            self._angle = 90
        elif 112.5 < angle <= 157.5:
            self._angle = 135
        elif 157.5 < angle <= 202.5:
            self._angle = 180
        elif 202.5 < angle <= 247.5:
            self._angle = 225
        elif 247.5 < angle <= 292.5:
            self._angle = 270
        elif 292.5 < angle <= 337.5:
            self._angle = 315

    def _createAnyAngleRuler(self, angle):
        self._angle = angle

    def _createRulerTicks(self):
        self._tickTuples = list()
        direction = QPoint(0, 0)
        perpendicular = QPoint(0, 0)
        if self._draftLine.length() >= self._tickGap:
            numberOfTicks = math.ceil(self._draftLine.length() / self._tickGap)
            direction = self._draftLine.p2() - self._draftLine.p1()
            if direction != QPoint(
                0, 0
            ):  # no need for a tick when the line is zero length
                direction /= direction.manhattanLength()
                perpendicular = QPointF(-direction.y(), direction.x())
                for i in range(numberOfTicks):
                    self._tickTuples.append(
                        ddef.rulerTuple(
                            self._draftLine.p1()
                            + i * self._tickGap * direction
                            + perpendicular * self._tickLength,
                            QLineF(
                                self._draftLine.p1() + i * self._tickGap * direction,
                                self._draftLine.p1()
                                + i * self._tickGap * direction
                                + perpendicular * self._tickLength,
                            ),
                            str(float(i)),
                        )
                    )
        self._tickTuples.append(
            ddef.rulerTuple(
                self.draftLine.p2() + direction * 2,
                QLineF(
                    self._draftLine.p2(),
                    self.draftLine.p2() + +perpendicular * self._tickLength,
                ),
                str(round(self._draftLine.length() / self._tickGap, 3)),
            )
        )
        point1 = self._draftLine.p1().toPoint()
        point2 = (
            self._draftLine.p2()
            + perpendicular
            * (self._tickLength + len(self._tickTuples[-1].text) * self._fm.maxWidth())
        ).toPoint()
        self._rect = QRectF(point1, point2).normalized()

    def boundingRect(self) -> QRectF:
        return self._rect.adjusted(-30, -10, 30, 10)

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(self._selectedPen)
            painter.drawRect(self.childrenBoundingRect())
        else:
            painter.setPen(self._pen)
        painter.drawLine(self._draftLine)
        painter.setFont(self._tickFont)
        for tickTuple in self._tickTuples:
            painter.drawLine(tickTuple.line)
            painter.save()
            painter.translate(tickTuple.point)
            painter.rotate(self.angle)
            painter.translate(-tickTuple.point)
            painter.drawText(tickTuple.point.x(), tickTuple.point.y(), tickTuple.text)
            painter.restore()

    @property
    def draftLine(self):
        return self._draftLine

    @draftLine.setter
    def draftLine(self, line: QLineF):
        self.prepareGeometryChange()
        self._draftLine = line
        angle = self._draftLine.angle()
        self._determineAngle(angle)
        self._createRulerTicks()

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width: float):
        self._width = width

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value: int):
        self._mode = value

    @property
    def tickFont(self):
        return self._tickFont

    @property
    def tickGap(self):
        return self._tickGap


class layoutLabel(layoutShape):
    labelAlignments = ["Left", "Center", "Right"]
    labelOrients = ["R0", "R90", "R180", "R270", "MX", "MX90", "MY", "MY90"]

    def __init__(
        self,
        start: QPoint,
        labelText: str,
        fontFamily: str,
        fontStyle: str,
        fontHeight: str,
        labelAlign: str,
        labelOrient: str,
        layer: ddef.layLayer,
    ):
        super().__init__()
        self._start = start
        self._labelText = labelText
        self._fontFamily = fontFamily
        self._fontStyle = fontStyle
        self._fontHeight = fontHeight
        self._labelAlign = labelAlign
        self._labelOrient = labelOrient
        self._layer = layer
        self.definePensBrushes()
        self.fontDefinition(fontFamily, fontStyle)
        self._labelOptions = QTextOption()
        if self._labelAlign == layoutLabel.labelAlignments[0]:
            self._labelOptions.setAlignment(Qt.AlignmentFlag.AlignLeft)
        elif self._labelAlign == layoutLabel.labelAlignments[1]:
            self._labelOptions.setAlignment(Qt.AlignmentFlag.AlignCenter)
        elif self._labelAlign == layoutLabel.labelAlignments[2]:
            self._labelOptions.setAlignment(Qt.AlignmentFlag.AlignRight)
        self.setOrient()

    def fontDefinition(self, fontFamily, fontStyle):
        self._labelFont = QFont(fontFamily)
        self._labelFont.setStyleName(fontStyle)
        # self._labelFont.setPointSize(int(float(self._labelHeight)))
        self._labelFont.setPointSize(int(float(self._fontHeight)))
        self._labelFont.setKerning(False)
        # self.setOpacity(1)
        self._fm = QFontMetrics(self._labelFont)
        self._rect = self._fm.boundingRect(self._labelText)

    def __repr__(self):
        return (
            f"layoutLabel({self._start}, {self._labelText}, {self._fontFamily}, "
            f"{self._fontStyle}, {self._fontHeight}, {self._labelAlign}, "
            f"{self._labelOrient}, {self._layer})"
        )

    def definePensBrushes(self):
        self._pen = QPen(self._layer.pcolor, self._layer.pwidth, self._layer.pstyle)
        self._bitmap = QBitmap.fromImage(
            QPixmap(self._layer.btexture).scaled(10, 10).toImage()
        )
        self._brush = QBrush(self._layer.bcolor, self._bitmap)
        self._selectedPen = QPen(QColor("yellow"), self._layer.pwidth, Qt.DashLine)
        self._selectedBrush = QBrush(QColor("yellow"), self._bitmap)

    def setOrient(self):
        self.setTransformOriginPoint(self.mapFromScene(self._start))
        if self._labelOrient == layoutLabel.labelOrients[0]:
            self.setRotation(0)
        elif self._labelOrient == layoutLabel.labelOrients[1]:
            self.setRotation(90)
        elif self._labelOrient == layoutLabel.labelOrients[2]:
            self.setRotation(180)
        elif self._labelOrient == layoutLabel.labelOrients[3]:
            self.setRotation(270)
        elif self._labelOrient == layoutLabel.labelOrients[4]:
            self.flip("x")
        elif self._labelOrient == layoutLabel.labelOrients[5]:
            self.flip("x")
            self.setRotation(90)
        elif self._labelOrient == layoutLabel.labelOrients[6]:
            self.flip("y")
            self.setRotation(90)

    def boundingRect(self):
        return (
            QRect(
                self._start.x(),
                self._start.y(),
                self._rect.width(),
                self._rect.height(),
            )
            .normalized()
            .adjusted(-2, -2, 2, 2)
        )  #

    def shape(self) -> QPainterPath:
        path = QPainterPath()
        path.addRect(self.boundingRect())
        return path

    def paint(self, painter, option, widget):
        self._labelFont.setPointSize(int(self._fontHeight))
        painter.setFont(self._labelFont)
        if self.isSelected():
            painter.setPen(self._selectedPen)
            painter.drawRect(self.boundingRect())
            self.setZValue(99)
        else:
            painter.setPen(self._pen)
            self.setZValue(self._layer.z)
        painter.drawText(
            QPoint(self._start.x(), self._start.y() + self._rect.height()),
            self._labelText,
        )
        painter.drawPoint(self._start)

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        if self._layer.selectable:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        else:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)

    def flip(self, direction: str):
        currentTransform = self.transform()
        newTransform = QTransform()
        if direction == "x":
            currentTransform = newTransform.scale(-1, 1) * currentTransform
        elif direction == "y":
            currentTransform = newTransform.scale(1, -1) * currentTransform
        self.setTransform(currentTransform)

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value: QPoint):
        self.prepareGeometryChange()
        self._start = value

    @property
    def labelText(self):
        return self._labelText

    @labelText.setter
    def labelText(self, value):
        self.prepareGeometryChange()
        self._labelText = value
        self._rect = self._fm.boundingRect(self._labelText)

    @property
    def labelFont(self):
        return self._labelFont

    @property
    def fontFamily(self) -> str:
        return self._labelFont.family()

    @fontFamily.setter
    def fontFamily(self, familyName):
        fontFamilies = QFontDatabase.families(QFontDatabase.Latin)
        fixedFamilies = [
            family for family in fontFamilies if QFontDatabase.isFixedPitch(family)
        ]
        if familyName in fixedFamilies:
            self._labelFont.setFamily(familyName)
        else:
            self.scene().logger.error(f"Not a valid font name: {familyName}")

    @property
    def fontStyle(self):
        return self._labelFont.styleName()

    @fontStyle.setter
    def fontStyle(self, value):
        if value in QFontDatabase.styles(self._labelFont.family()):
            self._labelFont.setStyleName(value)
        else:
            self.scene().logger.error(f"Not a valid font style: {value}")

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, value: ddef.layLayer):
        self.prepareGeometryChange()
        self._layer = value
        self.definePensBrushes()

    @property
    def fontHeight(self):
        return self._fontHeight

    @fontHeight.setter
    def fontHeight(self, value: float):
        self.prepareGeometryChange()
        self._fontHeight = value
        self._labelFont.setPointSize(int(float(self._fontHeight)))
        self._fm = QFontMetrics(self._labelFont)
        self._rect = self._fm.boundingRect(self._labelText)

    @property
    def labelAlign(self):
        return self._labelAlign

    @labelAlign.setter
    def labelAlign(self, value):
        self.prepareGeometryChange()
        self._labelAlign = value

    @property
    def labelOrient(self):
        return self._labelOrient

    @labelOrient.setter
    def labelOrient(self, value):
        self.prepareGeometryChange()
        self._labelOrient = value


class layoutPin(layoutShape):
    pinDirs = ["Input", "Output", "Inout"]
    pinTypes = ["Signal", "Ground", "Power", "Clock", "Digital", "Analog"]

    def __init__(
        self,
        start,
        end,
        pinName: str,
        pinDir: str,
        pinType: str,
        layer: ddef.layLayer,
    ):
        super().__init__()

        self._pinName = pinName
        self._pinDir = pinDir
        self._pinType = pinType
        self._connected = False  # True if the pin is connected to a net.
        self._rect = QRect(start, end).normalized()
        self._start = self._rect.topLeft()
        self._end = self._rect.bottomRight()
        self._layer = layer
        self._definePensBrushes()
        self._label = None
        self._stretchSide = None
        self._stretchPen = QPen(QColor("red"), self._layer.pwidth, Qt.SolidLine)

    def _definePensBrushes(self):
        self._pen = QPen(self._layer.pcolor, self._layer.pwidth, self._layer.pstyle)
        self._bitmap = QBitmap.fromImage(
            QPixmap(self._layer.btexture).scaled(10, 10).toImage()
        )
        self._brush = QBrush(self._layer.bcolor, self._bitmap)
        self._selectedPen = QPen(QColor("yellow"), self._layer.pwidth, Qt.DashLine)
        self._selectedBrush = QBrush(QColor("yellow"), self._bitmap)

    def __repr__(self):
        return (
            f"layoutPin({self._start}, {self._end}, {self._pinName}, {self._pinDir}, "
            f"{self._pinType}, {self._layer})"
        )

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(self._selectedPen)
            painter.setBrush(self._selectedBrush)
        else:
            painter.setPen(self._pen)
            painter.setBrush(self._brush)
        painter.drawRect(self._rect)

    def boundingRect(self):
        return self._rect.adjusted(-2, 2, 2, 2)

    @property
    def pinName(self):
        return self._pinName

    @pinName.setter
    def pinName(self, value):
        self._pinName = value

    @property
    def pinDir(self):
        return self._pinDir

    @pinDir.setter
    def pinDir(self, value):
        self._pinDir = value

    @property
    def pinType(self):
        return self._pinType

    @pinType.setter
    def pinType(self, value):
        self._pinType = value

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, value: ddef.layLayer):
        self.prepareGeometryChange()
        self._layer = value
        self._definePensBrushes()

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start: QPoint):
        self.prepareGeometryChange()
        self._rect = QRectF(start, self._end).normalized()
        self._start = self._rect.topLeft()

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end: QPoint):
        self.prepareGeometryChange()
        self._rect = QRectF(self._start, end).normalized()
        self._end = self._rect.bottomRight()

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value: layoutLabel):
        if isinstance(value, layoutLabel):
            self._label = value

        else:
            self.scene().logger.error("Not a Label")

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, rect: QRect):
        self.prepareGeometryChange()
        self._rect = rect

    @property
    def stretchSide(self):
        return self._stretchSide

    @stretchSide.setter
    def stretchSide(self, value: str):
        self.prepareGeometryChange()
        self._stretchSide = value

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        if self._layer.selectable:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
            eventPos = event.pos().toPoint()
            if self._stretch:
                self.setFlag(QGraphicsItem.ItemIsMovable, False)
                if eventPos.x() == self._rect.left():
                    if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                        self.setCursor(Qt.SizeHorCursor)
                        self._stretchSide = layoutRect.sides[0]
                elif eventPos.x() == self._rect.right():
                    if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                        self.setCursor(Qt.SizeHorCursor)
                        self._stretchSide = layoutRect.sides[1]
                elif eventPos.y() == self._rect.top():
                    if self._rect.left() <= eventPos.x() <= self._rect.right():
                        self.setCursor(Qt.SizeVerCursor)
                        self._stretchSide = layoutRect.sides[2]
                elif eventPos.y() == self._rect.bottom():
                    if self._rect.left() <= eventPos.x() <= self._rect.right():
                        self.setCursor(Qt.SizeVerCursor)
                        self._stretchSide = layoutRect.sides[3]
        else:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        eventPos = event.pos().toPoint()
        if self._stretch and self._stretchSide:
            self.prepareGeometryChange()
            if self.stretchSide == layoutRect.sides[0]:
                self.setCursor(Qt.SizeHorCursor)
                self.rect.setLeft(eventPos.x())
            elif self.stretchSide == layoutRect.sides[1]:
                self.setCursor(Qt.SizeHorCursor)
                self.rect.setRight(eventPos.x() - int(self._pen.width() / 2))
            elif self.stretchSide == layoutRect.sides[2]:
                self.setCursor(Qt.SizeVerCursor)
                self.rect.setTop(eventPos.y())
            elif self.stretchSide == layoutRect.sides[3]:
                self.setCursor(Qt.SizeVerCursor)
                self.rect.setBottom(eventPos.y() - int(self._pen.width() / 2))
            self.update()
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        super().mouseReleaseEvent(event)
        if self.stretch:
            self._stretch = False
            self._stretchSide = None
            self.setCursor(Qt.ArrowCursor)


class layoutVia(layoutShape):
    def __init__(
        self,
        start: QPoint,
        viaDefTuple: ddef.viaDefTuple,
        width: int,
        height: int,
    ):
        super().__init__()
        end = start + QPoint(width, height)
        self._rect = QRectF(start, end).normalized()
        self._start = self._rect.topLeft()
        self._end = self._rect.bottomRight()
        self._viaDefTuple = viaDefTuple
        self._layer = viaDefTuple.layer
        self._name = viaDefTuple.name
        self._type = viaDefTuple.type
        self._width = width
        self._height = height
        self._definePensBrushes()

    def _definePensBrushes(self):
        self._pen = QPen(self._layer.pcolor, self._layer.pwidth, self._layer.pstyle)
        self._bitmap = QBitmap.fromImage(
            QPixmap(self._layer.btexture).scaled(10, 10).toImage()
        )
        self._brush = QBrush(self._layer.bcolor, self._bitmap)
        self._selectedPen = QPen(QColor("yellow"), self._layer.pwidth, Qt.DashLine)
        self._selectedBrush = QBrush(QColor("yellow"), self._bitmap)

    def __repr__(self):
        return f"layoutVia({self._start}, {self._end}, {self._layer})"

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(self._selectedPen)
        else:
            painter.setPen(self._pen)
        painter.setBrush(self._brush)
        painter.drawRect(self._rect)
        painter.drawLine(self._rect.bottomLeft(), self._rect.topRight())
        painter.drawLine(self._rect.topLeft(), self._rect.bottomRight())

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        if self._layer.selectable:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        else:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)

    def boundingRect(self):
        return self._rect.normalized().adjusted(-2, -2, 2, 2)

    def shape(self) -> QPainterPath:
        path = QPainterPath()
        path.addRect(self.boundingRect())
        return path

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, value: ddef.layLayer):
        self.prepareGeometryChange()
        self._layer = value

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, rect: QRect):
        self.prepareGeometryChange()
        self._rect = rect

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start: QPoint):
        self.prepareGeometryChange()
        self._rect.moveTo(self.mapFromScene(start))

    @property
    def width(self):
        return self._rect.width()

    @width.setter
    def width(self, value: int):
        self.prepareGeometryChange()
        self._rect.setWidth(value)

    @property
    def height(self):
        return self._rect.height()

    @height.setter
    def height(self, value: int):
        self._rect.setHeight(value)

    @property
    def viaDefTuple(self):
        return self._viaDefTuple

    @viaDefTuple.setter
    def viaDefTuple(self, value: ddef.viaDefTuple):
        self.prepareGeometryChange()
        self._viaDefTuple = value
        self.layer = self._viaDefTuple.layer
        self._name = self._viaDefTuple.name
        self._definePensBrushes()
        self.update()


class layoutViaArray(layoutShape):
    def __init__(
        self,
        start: QPoint,
        via: layoutVia,
        spacing: float,
        xnum: int,
        ynum: int,
    ):
        super().__init__()
        self._start = start
        self._via = via  # prototype via
        self._xnum = xnum
        self._ynum = ynum
        self._spacing = spacing
        self._placeVias(via, xnum, ynum)
        self.setFiltersChildEvents(True)
        self.setHandlesChildEvents(True)
        self.setFlag(QGraphicsItem.ItemContainsChildrenInShape, True)
        self._selectedPen = QPen(QColor("yellow"), 1, Qt.DashLine)
        self._rect = self.childrenBoundingRect()

    def _placeVias(self, via, xnum, ynum):
        for childVia in self.childItems():
            self.scene().removeItem(childVia)
        for i, j in itertools.product(range(xnum), range(ynum)):
            item = layoutVia(
                QPoint(
                    self._start.x() + i * (self._spacing + via.width),
                    self._start.y() + j * (self._spacing + via.height),
                ),
                self._via.viaDefTuple,
                via.width,
                via.height,
            )
            item.setFlag(QGraphicsItem.ItemIsSelectable, False)
            item.setFlag(QGraphicsItem.ItemStacksBehindParent, True)
            item.setParentItem(self)

    def __repr__(self):
        return f"layoutViaArray({self._via}, {self._xnum}, {self._ynum})"

    def boundingRect(self) -> QRectF:
        return self.childrenBoundingRect()

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(self._selectedPen)
            painter.drawRect(self._rect)

    def shape(self) -> QPainterPath:
        path = QPainterPath()
        path.addRect(self._rect)
        return path

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start: QPoint):
        self.prepareGeometryChange()
        self._start = start

    @property
    def xnum(self) -> int:
        return self._xnum

    @xnum.setter
    def xnum(self, value: int):
        self.prepareGeometryChange()
        self._xnum = value
        self._rect = self.childrenBoundingRect()

    @property
    def ynum(self) -> int:
        return self._ynum

    @ynum.setter
    def ynum(self, value: int):
        self.prepareGeometryChange()
        self._ynum = value
        self._rect = self.childrenBoundingRect()

    @property
    def via(self):
        return self._via

    @via.setter
    def via(self, value: layoutVia):
        self.prepareGeometryChange()
        self._via = value

    @property
    def width(self):
        return self._via.width

    @width.setter
    def width(self, value: float):
        self.prepareGeometryChange()
        self._via.width = value
        for childVia in self.childItems():
            childVia.width = value
        self._rect = self.childrenBoundingRect()

    @property
    def height(self):
        return self._via.height

    @height.setter
    def height(self, value: float):
        self.prepareGeometryChange()
        self._via.height = value
        for childVia in self.childItems():
            childVia.height = value
        self._rect = self.childrenBoundingRect()

    @property
    def spacing(self) -> float:
        return self._spacing

    @spacing.setter
    def spacing(self, value: float):
        self.prepareGeometryChange()
        self._spacing = value
        self._placeVias(self._via, self._xnum, self._ynum)
        self._rect = self.childrenBoundingRect()

    @property
    def xnum(self) -> int:
        return self._xnum

    @xnum.setter
    def xnum(self, value: int):
        self._xnum = value
        self._placeVias(self._via, self._xnum, self._ynum)
        self._rect = self.childrenBoundingRect()

    @property
    def ynum(self) -> int:
        return self._ynum

    @ynum.setter
    def ynum(self, value: int):
        self._ynum = value
        self._placeVias(self._via, self._xnum, self._ynum)
        self._rect = self.childrenBoundingRect()

    @property
    def viaDefTuple(self):
        return self._via.viaDefTuple

    @viaDefTuple.setter
    def viaDefTuple(self, value: ddef.viaDefTuple):
        self._via.viaDefTuple = value
        self.prepareGeometryChange()
        for childVia in self.childItems():
            childVia.viaDefTuple = value
        self._rect = self.childrenBoundingRect()


class layoutPolygon(layoutShape):
    def __init__(self, points: list, layer: ddef.layLayer):
        super().__init__()
        self._points = points
        self._layer = layer
        self._definePensBrushes()
        self._polygon = QPolygonF(self._points)
        self._selectedCorner = None
        self._selectedCornerIndex = None

    def __repr__(self):
        return f"layoutPolygon({self._points}, {self._layer})"

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(self._selectedPen)
            painter.setBrush(self._selectedBrush)
            if self._selectedCorner is not None:
                painter.drawEllipse(self._selectedCorner, 5, 5)
        else:
            painter.setPen(self._pen)
            painter.setBrush(self._brush)
        painter.drawPolygon(self._polygon)

    def _definePensBrushes(self):
        self._pen = QPen(self._layer.pcolor, self._layer.pwidth, self._layer.pstyle)
        self._bitmap = QBitmap.fromImage(
            QPixmap(self._layer.btexture).scaled(10, 10).toImage()
        )
        self._brush = QBrush(self._layer.bcolor, self._bitmap)
        self._selectedPen = QPen(QColor("yellow"), self._layer.pwidth, Qt.DashLine)
        self._selectedBrush = QBrush(QColor("yellow"), self._bitmap)

    def boundingRect(self) -> QRectF:
        return self._polygon.boundingRect()

    @property
    def polygon(self):
        return self._polygon

    @property
    def points(self) -> list:
        return self._points

    @points.setter
    def points(self, value: list):
        self.prepareGeometryChange()
        self._points = value
        self._polygon = QPolygonF(self._points)

    def addPoint(self, point: QPoint):
        self.prepareGeometryChange()
        self._points.append(point)
        self._polygon = QPolygonF(self._points)

    @property
    def tempLastPoint(self):
        return self._points[-1]

    @tempLastPoint.setter
    def tempLastPoint(self, value: QPoint):
        self.prepareGeometryChange()
        self._polygon = QPolygonF([*self._points, value])

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, value: ddef.layLayer):
        self.prepareGeometryChange()
        self._layer = value
        self._definePensBrushes()

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        if self._layer.selectable:
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
            eventPos = event.pos().toPoint()
            if self._stretch:
                self.setFlag(QGraphicsItem.ItemIsMovable, False)
                for point in self._points:
                    if (
                        eventPos - point
                    ).manhattanLength() <= self.scene().snapDistance:
                        self._selectedCorner = point
                        self._selectedCornerIndex = self._points.index(point)
        else:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            self.setFlag(QGraphicsItem.ItemIsSelectable, False)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        eventPos = event.pos().toPoint()
        if self._stretch and self._selectedCorner:
            self._points[self._selectedCornerIndex] = eventPos
            self.points = self._points
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        if self.stretch:
            self._stretch = False
            self._stretchSide = None
            self.setCursor(Qt.ArrowCursor)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import json

import revedaEditor.common.shapes as shp
import revedaEditor.common.labels as lbl


class symbolAttribute(object):
    def __init__(self, name: str, definition: str):
        self._name = name
        self._definition = definition

    def __str__(self):
        return f"{self.name}: {self.definition}"

    def __repr__(self):
        return f"{type(self)}({self.name},{self.definition})"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        assert isinstance(value, str)
        self._name = value

    @property
    def definition(self):
        return self._definition

    @definition.setter
    def definition(self, value):
        assert isinstance(value, str)
        self._definition = value


class symbolEncoder(json.JSONEncoder):
    def default(self, item):
        match type(item):
            case shp.symbolRectangle:
                itemDict = {
                    "type": "rect",
                    "rect": item.rect.getCoords(),
                    "loc": (item.scenePos() - item.scene().origin).toTuple(),
                    "ang": item.angle,
                }
                return itemDict
            case shp.symbolLine:
                itemDict = {
                    "type": "line",
                    "st": item.start.toTuple(),
                    "end": item.end.toTuple(),
                    "loc": (item.scenePos() - item.scene().origin).toTuple(),
                    "ang": item.angle,
                }
                return itemDict
            case shp.symbolCircle:
                itemDict = {
                    "type": "circle",
                    "cen": item.centre.toTuple(),
                    "end": item.end.toTuple(),
                    "loc": (item.scenePos() - item.scene().origin).toTuple(),
                    "ang": item.angle,
                }
                return itemDict
            case shp.symbolPolygon:
                pointsList = [item.mapToScene(point).toTuple() for point in item.points]
                itemDict = {
                    "type": "polygon",
                    "ps": pointsList,
                }
                return itemDict
            case shp.symbolArc:
                itemDict = {
                    "type": "arc",
                    "st": item.start.toTuple(),
                    "end": item.end.toTuple(),
                    "loc": (item.scenePos() - item.scene().origin).toTuple(),
                    "ang": item.angle,
                }
                return itemDict
            case shp.symbolPin:
                itemDict = {
                    "type": "pin",
                    "st": item.start.toTuple(),
                    "nam": item.pinName,
                    "pd": item.pinDir,
                    "pt": item.pinType,
                    "loc": (item.scenePos() - item.scene().origin).toTuple(),
                    "ang": item.angle,
                }
                return itemDict
            case shp.text:
                itemDict = {
                    "type": "text",
                    "st": item.start.toTuple(),
                    "tc": item.textContent,
                    "ff": item.fontFamily,
                    "fs": item.fontStyle,
                    "th": item.textHeight,
                    "ta": item.textAlignment,
                    "to": item.textOrient,
                    "loc": (item.scenePos() - item.scene().origin).toTuple(),
                    "ang": item.angle,
                }
                return itemDict
            case lbl.symbolLabel:
                itemDict = {
                    "type": "label",
                    "st": item.start.toTuple(),
                    "nam": item.labelName,
                    "def": item.labelDefinition,  # label as entered
                    "txt": item.labelText,  # shown label
                    "val": item.labelValue,  # label value
                    "vis": item.labelVisible,  # label visibility
                    "lt": item.labelType,
                    "ht": item.labelHeight,
                    "al": item.labelAlign,
                    "or": item.labelOrient,
                    "use": item.labelUse,
                    "loc": (item.scenePos() - item.scene().origin).toTuple(),
                }
                return itemDict
            case symbolAttribute:
                itemDict = {
                    "type": "attr",
                    "nam": item.name,
                    "def": item.definition,
                }
                return itemDict



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)

# net class definition.
from PySide6.QtCore import (
    QPoint,
    Qt,
    QLineF,
    QRectF,
    QPointF,
)
from PySide6.QtGui import (
    QPen,
    QStaticText,
    QPainterPath,
    QFont,
)
from PySide6.QtWidgets import (
    QGraphicsLineItem,
    QGraphicsItem,
    QGraphicsPathItem,
    QGraphicsEllipseItem,
    QGraphicsSceneMouseEvent,
    QGraphicsSceneHoverEvent,
)
import pdk.schLayers as schlyr
import math
import itertools as itt
from typing import Union, Optional, NamedTuple


# class crossingDot(QGraphicsEllipseItem):
#     dotDiameter = schlyr.crossingDotDiameter
#
#     def __init__(self, point: QPoint):
#
#         self.point = point
#         self._radius = crossingDot.dotDiameter
#         super().__init__(point.x() - self._radius, point.y() - self._radius,
#                          2 * self._radius, 2 * self._radius)
#         self.setPen(schlyr.wirePen)
#         self.setBrush(schlyr.wireBrush)
#         self._name = None
#
#     def paint(self, painter, option, widget) -> None:
#         if self.isSelected():
#             painter.setPen(schlyr.selectedWirePen)
#             painter.setBrush(schlyr.selectedWireBrush)
#         else:
#             painter.setPen(schlyr.wirePen)
#             painter.setBrush(schlyr.wireBrush)
#         painter.drawEllipse(self.point, self._radius, self._radius)
#
#     def __repr__(self):
#         return f"crossingDot({self.point},{self._radius})"
#
#     def findNets(self) -> set["schematicNet"]:
#         if self.scene():
#             return {netItem for netItem in self.scene().items(self.sceneBoundingRect()) if
#                     isinstance(netItem, schematicNet)}
#         else:
#             return set()
#
#


class pointSelfIndex(NamedTuple):
    point: QPoint
    selfIndex: int


class schematicNet(QGraphicsItem):
    _netSnapLines: dict
    __slots__ = (
        "_mode",
        "_name",
        "_nameConflict",
        "_nameAdded",
        "_nameSet",
        "_highlighted",
        "_flightLinesSet",
        "_connectedNetsSet",
        "_netIndexTupleSet",
        "_scene" "_netSnapLines",
        "_stretch",
        "_stretchSide",
        "_nameFont",
        "_nameItem",
        "_draftLine",
        "_shapeRect",
        "_boundingRect",
        "_angle",
        "_transformOriginPoint",
    )

    def __init__(self, start: QPoint, end: QPoint, mode: int = 0):
        super().__init__()
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.ItemIsFocusable, True)
        # self.setFlag(QGraphicsItem.ItemSendsScenePositionChanges, True)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
        self.setAcceptHoverEvents(True)
        self._mode = 0
        self._name: str = ""
        self._nameConflict: bool = False
        self._nameAdded: bool = False
        self._nameSet: bool = False
        self._highlighted: bool = False
        self._flightLinesSet: set[schematicNet] = set()
        self._connectedNetsSet: set[schematicNet] = set()
        self._netSnapLines: dict = {}
        self._stretch: bool = False
        self._stretchSide: str = ""
        self._nameFont = QFont()
        self._nameFont.setPointSize(8)
        self._nameItem = QStaticText(self._name)
        self._draftLine = QLineF(start, end)
        self._transformOriginPoint = self._draftLine.p1()
        self._angle = None
        match self._mode:
            case 0:
                self._angle = 90 * math.floor(
                    ((self._draftLine.angle() + 45) % 360) / 90
                )
            case 1:
                self._angle = 45 * math.floor(
                    ((self._draftLine.angle() + 22.5) % 360) / 45
                )
            case 2:
                self._angle = self._draftLine.angle()
        self._draftLine.setAngle(0)
        self.setTransformOriginPoint(self._transformOriginPoint)
        self._shapeRect = QRectF(self._draftLine.p1(), self._draftLine.p2()).adjusted(
            -2, -2, 2, 2
        )
        self._boundingRect = self._shapeRect.adjusted(-8, -8, 8, 8)
        self.setRotation(-self._angle)

    @property
    def draftLine(self):
        return self._draftLine

    @draftLine.setter
    def draftLine(self, line: QLineF):
        self.prepareGeometryChange()
        self._draftLine = line
        self._transformOriginPoint = line.p1()
        match self._mode:
            case 0:
                self._angle = 90 * math.floor(
                    ((self._draftLine.angle() + 45) % 360) / 90
                )
            case 1:
                self._angle = 45 * math.floor(
                    ((self._draftLine.angle() + 22.5) % 360) / 45
                )
            case 2:
                self._angle = self._draftLine.angle()
        self._draftLine.setAngle(0)
        self.setTransformOriginPoint(self._transformOriginPoint)

        self._shapeRect = (
            QRectF(self._draftLine.p1(), self._draftLine.p2())
            .normalized()
            .adjusted(-2, -2, 2, 2)
        )
        self._boundingRect = self._shapeRect.adjusted(-8, -8, 8, 8)
        self.setRotation(-self._angle)

    def shape(self) -> QPainterPath:
        path = QPainterPath()
        path.addRect(self._shapeRect)
        return path

    def boundingRect(self):
        return self._boundingRect

    def paint(self, painter, option, widget=...):
        painter.setFont(self._nameFont)
        if self.isSelected():
            pen = schlyr.selectedWirePen
        elif self._stretch:
            pen = schlyr.stretchWirePen
        elif self._highlighted:
            pen = schlyr.hilightPen
        else:
            pen = schlyr.wirePen

        painter.setPen(pen)
        painter.drawLine(self._draftLine)
        # painter.drawEllipse(self._draftLine.p1(), 2, 2)
        if self._nameSet:
            painter.save()
            painter.translate(
                self._draftLine.center().x(), self._draftLine.center().y()
            )
            painter.rotate(self._angle)
            painter.drawStaticText(0, 0, self._nameItem)
            painter.restore()

    def sceneEvent(self, event):
        """
        Handle events related to the scene.

        Args:
            event: The event to handle.

        Returns:
            True if the event was handled successfully, False otherwise.
        """
        # Check if the current scene has the drawWire edit mode enabled
        if self.scene() and self.scene().editModes.drawWire:
            return False
        else:
            # Call the parent class's sceneEvent method to handle the event
            super().sceneEvent(event)
        return True

    def __repr__(self):
        return f"schematicNet({self.sceneEndPoints})"

    def itemChange(self, change, value):
        return super().itemChange(change, value)

    #
    def mousePressEvent(self, event: QGraphicsSceneMouseEvent):
        super().mousePressEvent(event)
        if self._stretch:
            eventPos = event.pos().toPoint()
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            if (
                eventPos - self._draftLine.p1().toPoint()
            ).manhattanLength() <= self.scene().snapDistance:
                self.setCursor(Qt.SizeHorCursor)
                self._stretchSide = "p1"
            elif (
                eventPos - self._draftLine.p2().toPoint()
            ).manhattanLength() <= self.scene().snapDistance:
                self.setCursor(Qt.SizeHorCursor)
                self._stretchSide = "p2"
            self.scene().stretchNet(self, self._stretchSide)
        # else:
        #     orthoNets = list(filter(self.isOrthogonal, self.findOverlapNets()))
        #     for otherNet in orthoNets:
        #         for selfEnd, otherEnd in itt.product(self.sceneEndPoints, otherNet.sceneEndPoints):
        #             if selfEnd == otherEnd:
        #                 if otherNet.sceneEndPoints.index(otherEnd) == 0:
        #                     otherNet._stretchSide = 'p1'
        #                 else:
        #                     otherNet._stretchSide = 'p2'
        #                 self.scene().stretchNet(otherNet, otherNet._stretchSide)
        #                 break

    def hoverEnterEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        """
        Override the hoverEnterEvent method of QGraphicsItem.

        Args:
            event (QGraphicsSceneHoverEvent): The hover event.

        Returns:
            None
        """
        super().hoverEnterEvent(event)
        # Check if highlightNets flag is set in the scene
        if self.scene().highlightNets:
            sceneNetsSet = self.scene().findSceneNetsSet() - {self}
            self._connectedNetsSet, _ = self.scene().traverseNets({self}, sceneNetsSet)
            # Highlight the connected netItems
            for netItem in self._connectedNetsSet:
                if not (netItem.nameAdded or netItem.nameSet):
                    netItem.nameAdded = True
                    netItem.name = self._name
                netItem.highlight()

            # Create flight lines and add them to the scene
            for netItem in self._connectedNetsSet:
                flightLine = netFlightLine(
                    self.mapToScene(self._draftLine.center()),
                    netItem.mapToScene(netItem.draftLine.center()),
                )
                self._flightLinesSet.add(flightLine)
                self.scene().addItem(flightLine)

    def hoverLeaveEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        super().hoverLeaveEvent(event)
        if self._highlighted:
            for flightLine in self._flightLinesSet:
                self.scene().removeItem(flightLine)
            self._flightLinesSet = set()
        [netItem.unhighlight() for netItem in self._connectedNetsSet]

    def isParallel(self, otherNet: "schematicNet") -> bool:
        return abs((self.angle - otherNet.angle) % 180) < 1

    def isOrthogonal(self, otherNet: "schematicNet") -> bool:
        return abs((self.angle - otherNet.angle - 90) % 180) < 1

    def notParallel(self, otherNet: "schematicNet") -> bool:
        return not self.isParallel(otherNet)

    def findOverlapNets(self) -> set["schematicNet"]:
        """
        Find all netItems in the scene that overlap with self.sceneShapeRect.

        Returns:
            set: A set of netItems that overlap with self.sceneShapeRect.
        """
        if self.scene():
            overlapNets = {
                netItem
                for netItem in self.scene().items(self.sceneShapeRect)
                if isinstance(netItem, schematicNet) and netItem is not self
            }
            if self._nameSet or self._nameAdded:
                for netItem in overlapNets:
                    netItem.name = self._name
                    netItem.nameSet = self._nameSet
                    netItem.nameAdded = self._nameAdded
        else:
            overlapNets = set()
        return overlapNets

    def mergeNets(self) -> tuple["schematicNet", "schematicNet"]:
        """
        Merges overlapping nets and returns the merged net.

        Returns:
            Optional[schematicNet]: The merged net if there are overlapping nets, otherwise returns self.
        """
        # Find other nets that overlap with self
        otherNets = self.findOverlapNets()

        # If there are other nets
        if otherNets:
            # Filter the other nets to find the parallel ones
            parallelNets = [
                netItem for netItem in otherNets if self.isParallel(netItem)
            ]

            # If there are parallel nets
            if parallelNets:
                # Create an initialRect variable and set it to self's sceneShapeRect
                initialRect = self.sceneShapeRect

                # Iterate over the parallel nets
                for netItem in parallelNets:
                    # Update the initialRect by uniting it with each parallel net's
                    # sceneShapeRect
                    initialRect = initialRect.united(netItem.sceneShapeRect)

                    # Remove each parallel net from the scene
                    if netItem.scene():
                        netItem.scene().removeItem(netItem)

                # Adjust the initialRect by 2 pixels on each side
                newNetPoints = initialRect.adjusted(2, 2, -2, -2)

                # Get the coordinates of the adjusted rectangle
                x1, y1, x2, y2 = newNetPoints.getCoords()

                # Create a new schematicNet with the snapped coordinates
                newNet = schematicNet(
                    self.snapToGrid(QPoint(x1, y1)), self.snapToGrid(QPoint(x2, y2))
                )

                # If self has a name, set the name and nameSet of the newNet
                if self._nameSet or self._nameAdded:
                    newNet.name = self._name
                    newNet.nameSet = self._nameSet
                    newNet.nameAdded = self._nameAdded

                return self, newNet  # original net, new net

        # If there are no other nets or no parallel nets, return self
        return self, self

    def snapToGrid(self, point: Union[QPoint, QPointF]) -> QPoint:
        if self.scene():
            return self.scene().snapToGrid(point, self.scene().snapTuple)
        else:
            return point

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if name != "":  # net name should not be an empty string
            self.prepareGeometryChange()
            self._name = name
            self._nameItem = QStaticText(self._name)

    @property
    def nameSet(self) -> bool:
        """
        Check if the name of the net is explicitly set.

        Returns:
            bool: The value of the 'nameSet' attribute.
        """
        return self._nameSet

    @nameSet.setter
    def nameSet(self, value: bool):
        """
        If the name of the net is explicitly set, set this attribute to True.

        Args:
            value (bool): The value to set for the 'nameSet' attribute.

        Raises:
            AssertionError: If the input value is not a boolean.
        """
        assert isinstance(value, bool)
        self._nameSet = value

    @property
    def nameAdded(self) -> bool:
        """
        Name added is true if net name is set due to a connected net or pin.
        """
        return self._nameAdded

    @nameAdded.setter
    def nameAdded(self, value: bool):
        assert isinstance(value, bool)
        self._nameAdded = value

    @property
    def nameConflict(self) -> bool:
        """
        If two different names are attempted to be set for the net.
        """
        return self._nameConflict

    @nameConflict.setter
    def nameConflict(self, value: bool):
        assert isinstance(value, bool)
        self._nameConflict = value

    @property
    def sceneEndPoints(self):
        return [
            self.mapToScene(self._draftLine.p1()).toPoint(),
            self.mapToScene(self._draftLine.p2()).toPoint(),
        ]

    def highlight(self):
        self._highlighted = True
        self.update()

    def unhighlight(self):
        self._highlighted = False
        self.update()

    @property
    def highlighted(self):
        return self._highlighted

    @highlighted.setter
    def highlighted(self, value: bool):
        self._highlighted = value

    @property
    def sceneShapeRect(self) -> QRectF:
        return self.mapRectToScene(self._shapeRect).normalized().toRect()

    @property
    def stretch(self) -> bool:
        return self._stretch

    @stretch.setter
    def stretch(self, value: bool):
        self._stretch = value

    @property
    def stretchSide(self) -> int:
        """
        The end where the net is stretched.
        """
        return self._stretchSide

    @stretchSide.setter
    def stretchSide(self, value: str):
        self.prepareGeometryChange()
        self._stretchSide = value

    @property
    def angle(self) -> float:
        return self._angle

    @angle.setter
    def angle(self, value: float):
        self._angle = value

    @property
    def stretch(self) -> bool:
        return self._stretch

    @stretch.setter
    def stretch(self, value: bool):
        self._stretch = value


class netFlightLine(QGraphicsPathItem):
    wireHighlightPen = QPen(
        schlyr.wireHilightLayer.pcolor,
        schlyr.wireHilightLayer.pwidth,
        schlyr.wireHilightLayer.pstyle,
    )

    def __init__(self, start: QPoint, end: QPoint):
        self._start = start
        self._end = end
        super().__init__()

    def __repr__(self):
        return f"netFlightLine({self.mapToScene(self._start)},{self.mapToScene(self._end)})"

    def paint(self, painter, option, widget) -> None:
        painter.setPen(netFlightLine.wireHighlightPen)
        line = QLineF(self._start, self._end)
        perpendicularLine = QLineF(
            line.center(), line.center() + QPointF(-line.dy(), line.dx())
        )
        perpendicularLine.setLength(100)

        path = QPainterPath()
        path.moveTo(self._start)
        path.quadTo(perpendicularLine.p2(), self._end)
        painter.drawPath(path)


class guideLine(QGraphicsLineItem):
    def __init__(self, start: QPoint, end: QPoint):
        self._start = start
        self._end = end
        super().__init__(QLineF(self._start, self._end))
        self.setPen(schlyr.guideLinePen)
        self._name: str = ""
        self._nameSet: bool = False

    @property
    def sceneEndPoints(self) -> list[QPoint]:
        return [
            self.mapToScene(self.line().p1()).toPoint(),
            self.mapToScene(self.line().p2()).toPoint(),
        ]

    def __repr__(self):
        return f"guideLine({self.mapToScene(self.line().p1()), self.mapToScene(self.line().p2())}"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if name != "":  # net name should not be an empty string
            self._name = name

    @property
    def nameSet(self) -> bool:
        return self._nameSet

    @nameSet.setter
    def nameSet(self, value: bool):
        assert isinstance(value, bool)
        self._nameSet = value



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
import pathlib


class verilogaC:
    """
    This class represents an imported verilog-A module.
    """

    def __init__(self, pathObj: pathlib.Path):
        """
        Initialize the class with a given path object.

        Args:
            pathObj (pathlib.Path): The path object representing the file to be processed.
        """
        self._pathObj = pathObj
        self._vaModule = ""
        self.instanceParams = dict()
        self.modelParams = dict()
        self.pins = list()
        self.inPins = list()
        self.inoutPins = list()
        self.outPins = list()
        self._netlistLine = ""
        self.statementLines = list()
        self._pinOrder = ""

        with open(self._pathObj) as f:
            self.fileLines = f.readlines()

        self.findPinsParams(self.oneLiners(self.stripComments()))

    def stripComments(self) -> list:
        """
        Strip comments from the file lines and store the non-comment lines in the statementLines list.
        """

        statementLines = []
        comment = False
        for line in self.fileLines:
            # Concatenate the lines until it reaches a ';'
            stripLine = line.strip()

            if stripLine.startswith("/*"):
                # Set comment to True if the line starts with '/*'
                comment = True

            if not comment:
                doubleSlashLoc = stripLine.find("//")
                if doubleSlashLoc > -1:
                    # Remove single-line comments starting from '//'
                    stripLine = stripLine[:doubleSlashLoc]

                if stripLine:
                    # Add non-empty lines to statementLines list
                    statementLines.append(stripLine)

            if comment and stripLine.endswith("*/"):
                # Set comment to False if the line ends with '*/'
                comment = False

        return statementLines

    def oneLiners(self, statementLines: list):
        """
        Convert the statement lines into one-liners.
        """
        tempLine = ""
        oneLiners = list()
        for line in statementLines:
            stripLine = line.strip()
            if not stripLine.startswith("`include"):
                tempLine = f"{tempLine} {stripLine}"
                if tempLine.endswith(";"):
                    oneLiners.append(tempLine.strip())
                    tempLine = ""
        return oneLiners

    def findPinsParams(self, filteredLines: list):
        for line in filteredLines:
            splitLine = line.strip().split(" ", 1)
            match splitLine[0].lower():
                case "module":
                    if "(" in splitLine[1]:
                        self._vaModule = splitLine[1].split("(")[0]
                    else:
                        self._vaModule = splitLine[1].replace(";", "").strip()
                    if "(" in splitLine[1] and ")" in splitLine[1]:
                        pinList = line.split("(")[1].split(")")[0].split(",")
                        self.pins = [pin.strip() for pin in pinList]
                    else:
                        self.pins = []
                case "in":
                    rawPins = line.replace("in ", "").replace(";", "").split(",")
                    self.inPins.extend([pin.strip() for pin in rawPins])
                case "input":
                    rawPins = line.replace("input ", "").replace(";", "").split(",")
                    self.inPins.extend([pin.strip() for pin in rawPins])
                case "out":
                    rawPins = line.replace("out ", "").replace(";", "").split(",")
                    self.outPins.extend([pin.strip() for pin in rawPins])
                case "output":
                    rawPins = line.replace("output ", "").replace(";", "").split(",")
                    self.outPins.extend([pin.strip() for pin in rawPins])
                case "inout":
                    rawPins = line.replace("inout ", "").replace(";", "").split(",")
                    self.inoutPins.extend([pin.strip() for pin in rawPins])
                case "parameter":
                    paramDefPart = (
                        line.replace("parameter", "").replace(";", "").split("*(")[0]
                    )
                    paramName = paramDefPart.split("=")[0].split()[-1].strip()
                    paramValue = paramDefPart.split("=")[1].split()[0].strip()
                    if "(*" in line:
                        paramAttr = line.strip().split("(*")[1]
                    else:
                        paramAttr = ""
                    if "type" in paramAttr and '"instance"' in paramAttr:
                        # parameter value is between = and (*
                        self.instanceParams[paramName] = paramValue
                        if "xyceAlsoModel" in paramAttr and '"yes"' in paramAttr:
                            self.modelParams[paramName] = paramValue
                    else:  # no parameter attribute statement
                        self.modelParams[paramName] = paramValue

    @property
    def pathObj(self):
        return self._pathObj

    @pathObj.setter
    def pathObj(self, value: pathlib.Path):
        assert isinstance(value, pathlib.Path)
        self._pathObj = value

    @property
    def pinOrder(self):
        return self._pinOrder

    @pinOrder.setter
    def pinOrder(self, value: str):
        assert isinstance(value, str)
        self._pinOrder = value

    @property
    def netlistLine(self):
        # Join the pins with commas
        self._pinOrder = ",".join(self.pins)

        # Create a string of instance parameters in the format [@key:key=%:key=item]
        instParamString = " ".join(
            [
                f"[@{key}:{key}=%:{key}={item}]"
                for key, item in self.instanceParams.items()
            ]
        )

        # Create the netlist line string with the formatted values
        self._netlistLine = f"Y{self._vaModule} @instName @pinList  {self._vaModule}Model {instParamString}"

        # Return the netlist line
        return self._netlistLine

    @property
    def vaModule(self):
        return self._vaModule


class spiceC:
    def __init__(self, pathObj: pathlib.Path):
        self._pathObj = pathObj
        self._statementLines = list()
        self._pinOrder = ""
        with self._pathObj.open("r") as f:
            self._fileLines = f.readlines()
        self.subcktParams = self.extractSubcktParams()
        self._netlistLine: str = ""

    @property
    def pathObj(self):
        return self._pathObj

    @pathObj.setter
    def pathObj(self, value: pathlib.Path):
        assert isinstance(value, pathlib.Path)
        self._pathObj = value

    @property
    def pinOrder(self):
        return self._pinOrder

    @pinOrder.setter
    def pinOrder(self, value: str):
        assert isinstance(value, str)
        self._pinOrder = value

    @property
    def netlistLine(self):
        # Create a string of instance parameters in the format [@key:key=%:key=item]
        instParamString = " ".join(
            [
                f"[@{key}:{key}=%:{key}={item}]"
                for key, item in self.subcktParams["params"].items()
            ]
        )
        if instParamString.strip():
            # Create the netlist line string with the formatted values
            self._netlistLine = f'X@instName @pinList {self.subcktParams["name"]} PARAM: {instParamString}'
        else:
            self._netlistLine = f'X@instName @pinList {self.subcktParams["name"]}'

        # Return the netlist line
        return self._netlistLine

    def subcktLineExtract(self):
        """ """
        subcktLines = ""
        subcktStart = None
        increment = 1

        # Iterate over each line in the file
        for linenumber, line in enumerate(self._fileLines):
            # Check if the line starts with '.SUBCKT'
            if line.upper().startswith(".SUBCKT"):
                subcktStart = linenumber
                subcktLines = f"{line.strip()}"

            # Check if the line starts with '+' and is the next line after the subcircuit start line
            if (
                subcktLines
                and linenumber == subcktStart + increment
                and line.startswith("+")
            ):
                subcktLines = f"{subcktLines} {line[1:].strip()}"
                increment += 1
        return subcktLines

    def extractSubcktParams(self):
        """ """
        subcktDict = dict()
        subcktDict.setdefault("params", dict())
        subcktLine = self.subcktLineExtract()
        tokens = subcktLine.split()
        subcktDict["name"] = tokens[1]
        capSubcktLine = subcktLine.upper()
        if "PARAM:" in capSubcktLine:
            subcktDict["pins"] = tokens[2: capSubcktLine.split().index("PARAM:")]
            paramsStringList = tokens[capSubcktLine.split().index("PARAM:") + 1 :]
            for index, paramString in enumerate(paramsStringList):
                if paramString.strip() == "=":
                    subcktDict["params"][
                        paramsStringList[index - 1].strip()
                    ] = paramsStringList[index + 1].strip()
        else:
            subcktDict["pins"] = tokens[2:]
        self._pinOrder = ",".join(subcktDict["pins"])
        return subcktDict
